// Copyright (c) 2010, 2011 Paul Pogonyshev.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


#undef  MCT_SELF_VALIDATION
#define MCT_SELF_VALIDATION     1

#include "tests/map-common.hpp"
#include "tests/map-parameters.hpp"


BOOST_AUTO_TEST_SUITE (map_main)


BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_map_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;

  BOOST_CHECK       (map.empty ());
  BOOST_CHECK_EQUAL (map.size (), 0u);
  BOOST_CHECK       (!map.uses_dynamic_memory ());
  BOOST_CHECK_EQUAL (map.used_memory (), sizeof (map_base_type));
  BOOST_CHECK_EQUAL (map.begin  (), map.end  ());
  BOOST_CHECK_EQUAL (map.cbegin (), map.cend ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_map_explicit_size_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map (1000);

  BOOST_CHECK (!map.uses_dynamic_memory ());

  map.insert (make_pair (key_data::generate (0), mapped_data::generate (0)));

  BOOST_CHECK    (map.uses_dynamic_memory ());
  BOOST_CHECK_GE (map.bucket_count (), 1000);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_map_copy_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;
  map_type  map2 (map);

  BOOST_CHECK       (map2.empty ());
  BOOST_CHECK_EQUAL (map2.size (), 0u);
  BOOST_CHECK       (!map2.uses_dynamic_memory ());
  BOOST_CHECK_EQUAL (map2.begin  (), map2.end  ());
  BOOST_CHECK_EQUAL (map2.cbegin (), map2.cend ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_map_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;
  map_type  map2;

  map = map2;

  BOOST_CHECK       (map.empty ());
  BOOST_CHECK_EQUAL (map.size (), 0u);
  BOOST_CHECK       (!map.uses_dynamic_memory ());
  BOOST_CHECK_EQUAL (map.begin  (), map.end  ());
  BOOST_CHECK_EQUAL (map.cbegin (), map.cend ());

  map.insert (RANGE (value_data::values1 ()));
  map = map2;

  // Check the same invariants if assigning over a non-empty map.
  BOOST_CHECK       (map.empty ());
  BOOST_CHECK_EQUAL (map.size (), 0u);
  BOOST_CHECK       (!map.uses_dynamic_memory ());
  BOOST_CHECK_EQUAL (map.begin  (), map.end  ());
  BOOST_CHECK_EQUAL (map.cbegin (), map.cend ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_map_erase_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  map_type  map;

  for (typename vector <key_type>::const_iterator scan = key_data::values1 ().begin ();
       scan != key_data::values1 ().end (); ++scan)
    BOOST_CHECK_EQUAL (map.erase (*scan), 0u);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_map_rehash_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;

  BOOST_CHECK (!map.uses_dynamic_memory ());

  map.rehash (1000);

  BOOST_CHECK    (!map.uses_dynamic_memory ());
  BOOST_CHECK_GE (map.bucket_count (), 1000);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_map_rehash_cleared_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map (RANGE (value_data::values1 ()));
  map.clear ();

  BOOST_CHECK (map.uses_dynamic_memory ());

  map.rehash (2 * map.bucket_count ());

  BOOST_CHECK (!map.uses_dynamic_memory ());
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_range_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map (RANGE (value_data::values1 ()));
  if (linked)
    assert_identical_order (map, value_data::keys1 ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_range_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;
  map_type  map (value_data::values1 ().begin (), value_data::values1 ().begin ());
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_initializer_constructor_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int, string>::map_type  map_type;

  map_type  map1 { { 1, "foo" }, { 2, "bar" }, { 3, "baz" }, { 4, "ham" } };
  map_type  map2 { { 1, "foo" }, { 3, "bar" }, { 1, "baz" }, { 3, "ham" }, { 2, "spam" },
                   { 2, "egg" }, { 2, "bop" } };

  map_type  expected { { 1, "foo" }, { 3, "bar" }, { 2, "spam" } };
  BOOST_CHECK_EQUAL (map2, expected);

  if (implementation::linked)
    {
      assert_identical_order (map1, vector <int> { 1, 2, 3, 4 });
      assert_identical_order (map2, vector <int> { 1, 3, 1, 3, 2, 2, 2 });
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_initializer_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;
  map_type  map ({ });
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_find_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map (RANGE (value_data::values1 ()));

  for (typename vector <key_type>::const_iterator key_scan = value_data::keys1 ().begin ();
       key_scan != value_data::keys1 ().end (); ++key_scan)
    BOOST_CHECK (contains (map, *key_scan));

  for (typename vector <key_type>::const_iterator key_scan = key_data::values2 ().begin ();
       key_scan != key_data::values2 ().end (); ++key_scan)
    {
      BOOST_CHECK_EQUAL (contains (map, *key_scan),
                         (find (RANGE (value_data::keys1 ()), *key_scan)
                          != range_end (value_data::keys1 ())));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_count_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map (RANGE (value_data::values1 ()));

  for (typename vector <key_type>::const_iterator key_scan = value_data::keys1 ().begin ();
       key_scan != value_data::keys1 ().end (); ++key_scan)
    BOOST_CHECK_EQUAL (map.count (*key_scan), 1u);

  for (typename vector <key_type>::const_iterator key_scan = key_data::values2 ().begin ();
       key_scan != key_data::values2 ().end (); ++key_scan)
    {
      BOOST_CHECK_EQUAL (map.count (*key_scan),
                         ((find (RANGE (value_data::keys1 ()), *key_scan)
                           != range_end (value_data::keys1 ()))
                          ? 1u : 0u));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_equal_range_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map (RANGE (value_data::values1 ()));

  for (typename vector <key_type>::const_iterator key_scan = value_data::keys1 ().begin ();
       key_scan != value_data::keys1 ().end (); ++key_scan)
    {
      pair <const_iterator, const_iterator>  range (map.equal_range (*key_scan));
      BOOST_CHECK       (map.key_eq () (range.first->first, *key_scan));
      BOOST_CHECK_EQUAL (++range.first, range.second);
    }

  for (typename vector <key_type>::const_iterator key_scan = key_data::values2 ().begin ();
       key_scan != key_data::values2 ().end (); ++key_scan)
    {
      pair <const_iterator, const_iterator>  range (map.equal_range (*key_scan));

      if (find (RANGE (value_data::keys1 ()), *key_scan) == range_end (value_data::keys1 ()))
        BOOST_CHECK_EQUAL (range.first, range.second);
      else
        {
          BOOST_CHECK       (map.key_eq () (range.first->first, *key_scan));
          BOOST_CHECK_EQUAL (++range.first, range.second);
        }
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;
  for (typename vector <key_type>::const_iterator key_scan = key_data::values1 ().begin ();
       key_scan != key_data::values1 ().end (); ++key_scan)
    map.insert (make_pair (*key_scan, mapped_type ()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_range_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;
  map.insert (RANGE (value_data::values1 ()));
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_moving_1,
                               implementation, test_implementations)
{
  // Note that we use 'pair <const ..., ...>' in this test.  While inserting a pair
  // without constified first member works fine, this involves some hidden conversions and
  // those call copy or move constructors, thus making the test fail.  We only want to
  // test whether the containers themselves don't copy/move when they shouldn't, so the
  // usage is a little bit cumbersome.

  // FIXME: This part of the test is currently disabled, because it is not expected to
  //        pass.  The problem is that 'value_type' in a map is 'pair <const key_type,
  //        mapped_type>' and constness of the first field prevents move constructor from
  //        kicking in.  I guess it can be solved and eventually should be, but this is
  //        not very important.
  if (false)
  {
    typedef  typename implementation::template resolve <must_not_be_copied, int>::map_type
             map_type;

    // Make it large, so it never rehashes and thus copies elements.
    map_type  map1 (1000);

    for (int k = 0; k < 100; ++k)
      {
        map1.insert (pair <const must_not_be_copied, int> (must_not_be_copied (k), 0));
        map1.validate_integrity ();
      }

    map_type  map2 (1000);

    for (int k = 0; k < 100; ++k)
      {
        map2[must_not_be_copied (k)] = 0;
        map2.validate_integrity ();
      }
  }

  {
    typedef  typename implementation::template resolve <int, must_not_be_copied>::map_type
             map_type;

    map_type  map1 (1000);

    for (int k = 0; k < 100; ++k)
      {
        map1.insert (pair <const int, must_not_be_copied> (k, must_not_be_copied (k)));
        map1.validate_integrity ();
      }
  }

  {
    typedef  typename implementation::template resolve <must_not_be_moved, int>::map_type
             map_type;

    map_type  map1 (1000);

    for (int k = 0; k < 100; ++k)
      {
        must_not_be_moved                    key (k);
        pair <const must_not_be_moved, int>  value (key, 0);

        map1.insert (value);
        map1.validate_integrity ();
      }

    map_type  map2 (1000);

    for (int k = 0; k < 100; ++k)
      {
        must_not_be_moved  key (k);

        map2[key] = 0;
        map2.validate_integrity ();
      }
  }

  {
    typedef  typename implementation::template resolve <int, must_not_be_moved>::map_type
             map_type;

    map_type  map1 (1000);

    for (int k = 0; k < 100; ++k)
      {
        must_not_be_moved                    mapped (k);
        pair <const int, must_not_be_moved>  value (k, mapped);

        map1.insert (value);
        map1.validate_integrity ();
      }
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_initializers_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int, string>::map_type     map_type;
  typedef  typename implementation::template resolve <int, string>::robust_type  robust_type;

  map_type  map { { 0, "foo" }, { 1, "bar" }, { 2, "baz" } };
  map.insert ({ { 3, "ham" }, { 4, "spam" }, { 5, "egg" } });

  vector <pair <int, string> >  expected_vector { { 0, "foo" }, { 1, "bar"  }, { 2, "baz" },
                                                  { 3, "ham" }, { 4, "spam" }, { 5, "egg" } };
  robust_type                   expected (expected_vector.begin (), expected_vector.end ());

  map.validate_integrity ();
  BOOST_CHECK_EQUAL (map, expected);

  if (implementation::linked)
    assert_identical_order (map, keys_of (expected));
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_emplace_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int_wrapper, int>::map_type  map_type;

  map_type  map;
  for (int k = 0; k < 50; ++k)
    {
      map.emplace (make_pair (k, k));
      map.validate_integrity ();
    }

  BOOST_CHECK_EQUAL (map.size (), 50u);
  if (implementation::linked)
    assert_identical_order (map, create_vector (int_wrapper::wrap, map.size ()));
}

#endif


BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_erase_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  map_type  map;

  for (typename vector <assignable_value_type>::const_iterator value_scan
         = value_data::values1 ().begin ();
       value_scan != value_data::values1 ().end (); ++value_scan)
    {
      // Simply insert and immediately erase one element.
      map.insert (*value_scan);
      BOOST_CHECK (map.erase (value_scan->first));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_copy_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map1 (RANGE (value_data::values1 ()));
  map_type  map2 (RANGE (value_data::values2 ()));
  map_type  map3 (map1);
  map_type  map4 (map2);

  BOOST_CHECK_EQUAL (map3, map1);
  BOOST_CHECK_EQUAL (map4, map2);

  if (linked)
    {
      assert_identical_order (map3, keys_of (map1));
      assert_identical_order (map4, keys_of (map2));
    }
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_move_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map (move (map_type (RANGE (value_data::values1 ()))));
  BOOST_CHECK_EQUAL (map, map_type (RANGE (value_data::values1 ())));
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map1 (RANGE (value_data::values1 ()));
  map_type  map2;

  BOOST_CHECK_EQUAL (map2.size (), 0u);

  map2 = map1;
  BOOST_CHECK_EQUAL (map2, map1);

  if (linked)
    assert_identical_order (map1, keys_of (map2));
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_self_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map1 (RANGE (value_data::values1 ()));
  map_type  map2 (RANGE (value_data::values1 ()));

  BOOST_CHECK_EQUAL (map1, map2);
  if (linked)
    assert_identical_order (map1, keys_of (map2));

  map1 = map1;

  BOOST_CHECK_EQUAL (map1, map2);
  if (linked)
    assert_identical_order (map1, keys_of (map2));
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_move_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;

  map = move (map_type (RANGE (value_data::values1 ())));
  BOOST_CHECK_EQUAL (map, map_type (RANGE (value_data::values1 ())));
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_initializer_assignment_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int, string>::map_type  map_type;

  map_type  map { { 1, "foo" }, { 2, "bar" }, { 3, "baz" }, { 4, "bop" } };
  map = { { 4, "ham" }, { 5, "spam" }, { 6, "egg" } };

  if (implementation::linked)
    assert_identical_order (map, vector <int> { 4, 5, 6 });
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_swap_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map1 (RANGE (value_data::values1 ()));
  map_type  map2 (RANGE (value_data::values2 ()));
  map_type  map3 (RANGE (value_data::values1 ()));
  map_type  map4 (RANGE (value_data::values2 ()));

  BOOST_CHECK_EQUAL (map1, map3);
  BOOST_CHECK_EQUAL (map2, map4);

  if (linked)
    {
      assert_identical_order (map1, keys_of (map3));
      assert_identical_order (map2, keys_of (map4));
    }

  swap (map3, map4);

  BOOST_CHECK_EQUAL (map1, map4);
  BOOST_CHECK_EQUAL (map2, map3);

  if (linked)
    {
      assert_identical_order (map1, keys_of (map4));
      assert_identical_order (map2, keys_of (map3));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_clear_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map (RANGE (value_data::values1 ()));
  map.clear ();
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_value_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  map_type  map1 (RANGE (value_data::values1 ()));
  map_type  map2 (range_begin (value_data::values1 ()) + 1, range_end (value_data::values1 ()));

  // Make sure funny comparison functions don't ruin this test.
  if (map1.size () != map2.size ())
    {
      BOOST_CHECK       (map1.erase (value_data::values1 () [0].first));
      BOOST_CHECK_EQUAL (map1, map2);

      if (linked)
        assert_identical_order (map1, keys_of (map2));
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_position_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  map_type  map (RANGE (value_data::values1 ()));

  const size_t  num_passes_expected = map.size ();
  size_t        num_passes          = 0;

  for (iterator scan = map.begin (), end = map.end (); scan != end; ++num_passes)
    scan = map.erase (scan);

  BOOST_CHECK       (map.empty ());
  BOOST_CHECK_EQUAL (num_passes, num_passes_expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_range_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  map_type  map (RANGE (value_data::values1 ()));

  map.erase (map.begin (), map.end ());
  BOOST_CHECK (map.empty ());
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_subscription_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map;
  int       k = 0;
  for (typename vector <key_type>::const_iterator key_scan = key_data::values1 ().begin ();
       key_scan != key_data::values1 ().end (); ++key_scan)
    map.subscript (*key_scan, mapped_data::generate (k++));
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_comparison_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  map_type  map1 (RANGE (value_data::values1 ()));
  map_type  map2 (RANGE (value_data::values1 ()));
  map_type  map3 (RANGE (value_data::values2 ()));
  map_type  map4 (RANGE (value_data::values2 ()));

  BOOST_CHECK_EQUAL (map1, map2);
  BOOST_CHECK_EQUAL (map3, map4);
  BOOST_CHECK_NE    (map1, map3);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_growth_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  // Note: fairly small, because testing after each step is slow.
  map_type  map;
  for (int k = 0; k < 200; ++k)
    map.insert (make_pair (key_data::generate (k), mapped_data::generate (k)));

  BOOST_CHECK_EQUAL (map.size (), 200u);
  if (linked)
    assert_identical_order (map, create_vector (key_data::generate, map.size ()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_shrink_1,
                               parameters, test_parameters_std_equal_normal)
{
  COMMON_TEST_SETUP;

  const size_t                    num_items = 200;
  vector <assignable_value_type>  items (create_vector (value_data::generate, num_items));
  map_type                        map   (items.begin (), items.end ());

  BOOST_CHECK_EQUAL (map.size (), num_items);

  const_iterator  almost_end = map.end ();
  for (int k = 0; k < 10; ++k)
    --almost_end;

  map.erase (map.begin (), almost_end);

  BOOST_CHECK_EQUAL (map.size (), 10u);
  if (linked)
    assert_identical_order (map, create_vector (key_data::generate, num_items, num_items - 10));

  map.rehash (32);

  BOOST_CHECK_EQUAL (map.bucket_count (), 32u);
  if (linked)
    assert_identical_order (map, create_vector (key_data::generate, num_items, num_items - 10));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_tiny_map_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  static  const float  load_factors[] = { 0.01, 0.1, 0.25, 0.5, 0.75, 0.9, 0.99 };

  for (const float* scan = range_begin (load_factors); scan != range_end (load_factors); ++scan)
    {
      BOOST_TEST_CHECKPOINT ("testing with maximum load factor " << *scan);

      map_type  map (1);
      map.max_load_factor (*scan);

      for (int k = 0; k < 50; ++k)
        map.insert (value_data::generate (k));

      if (linked)
        assert_identical_order (map, create_vector (key_data::generate, 50));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_max_load_factor_decrease_1,
                               implementation, test_implementations)
{
  typedef  test_map <typename implementation::template resolve <int, int>::map_type>  map_type;

  map_type  map;

  // Make sure 'map' is full for the default maximum load factor.
  for (int k = 0; ; k++)
    {
      map_type  copy (map);
      copy.insert (make_pair (k, k));

      if (copy.bucket_count () == map.bucket_count ())
        map.swap (copy);
      else
        break;
    }

  // Just something very small; implementation will clamp to an acceptable value.
  map.max_load_factor (0.0f);
}



BOOST_AUTO_TEST_CASE_TEMPLATE (test_intrusive_bucket_size_1,
                               implementation, test_implementations)
{
  {
    typedef  test_map <typename implementation::template resolve <ham, int>::map_type>  map_type;

    map_type  map;

    BOOST_CHECK       (!map_type::bucket_type::KEEPS_HASHES);
    BOOST_CHECK_EQUAL (sizeof (typename map_type::iterator), sizeof (pair <ham, int>*));
  }

  {
    typedef  test_map <typename implementation::template resolve <int, ham>::map_type>  map_type;

    map_type  map;

    BOOST_CHECK       (!map_type::bucket_type::KEEPS_HASHES);
    BOOST_CHECK_EQUAL (sizeof (typename map_type::iterator), sizeof (pair <int, ham>*));
  }

  {
    typedef  test_map <typename implementation::template resolve <ham, ham>::map_type>  map_type;

    map_type  map;

    BOOST_CHECK       (!map_type::bucket_type::KEEPS_HASHES);
    BOOST_CHECK_EQUAL (sizeof (typename map_type::iterator), sizeof (pair <ham, ham>*));
  }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_cross_class_comparison_1,
                               implementation, test_implementations)
{
  typedef  test_map <typename implementation::template resolve <int, int>::map_type>  map_type;

  vector <pair <int, int> >  values1 (create_vector (Data <pair <int, int> >::generate, 10));
  vector <pair <int, int> >  values2 (values1);

  values2[3].second = 100;

  map_type  map1 (RANGE (values1));
  map_type  map2 (RANGE (values2));

  // Damn macros, can't have commas in arguments.
  typedef  closed_hash_map <int, int>  closed_hash_map_int;
  BOOST_CHECK_EQUAL (map1, closed_hash_map_int (RANGE (values1)));
  BOOST_CHECK_NE    (map1, closed_hash_map_int (RANGE (values2)));

  typedef  linked_hash_map <int, int>  linked_hash_map_int;
  BOOST_CHECK_EQUAL (map1, linked_hash_map_int (RANGE (values1)));
  BOOST_CHECK_NE    (map1, linked_hash_map_int (RANGE (values2)));

  typedef  huge_linked_hash_map <int, int>  huge_linked_hash_map_int;
  BOOST_CHECK_EQUAL (map1, huge_linked_hash_map_int (RANGE (values1)));
  BOOST_CHECK_NE    (map1, huge_linked_hash_map_int (RANGE (values2)));

  typedef  forward_hash_map <int, int>  forward_hash_map_int;
  BOOST_CHECK_EQUAL (map1, forward_hash_map_int (RANGE (values1)));
  BOOST_CHECK_NE    (map1, forward_hash_map_int (RANGE (values2)));

  typedef  huge_forward_hash_map <int, int>  huge_forward_hash_map_int;
  BOOST_CHECK_EQUAL (map1, huge_forward_hash_map_int (RANGE (values1)));
  BOOST_CHECK_NE    (map1, huge_forward_hash_map_int (RANGE (values2)));
}


BOOST_AUTO_TEST_SUITE_END ()


// Local variables:
// mode: c++
// c-basic-offset: 2
// indent-tabs-mode: nil
// fill-column: 90
// End:
