// Copyright (c) 2009, 2010, 2011 Paul Pogonyshev.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


// Even though we perform our own integrity validation, and even comparisons to a known
// good implementation in our tests, we still let the tables validate themselves with
// MCT_SELF_VALIDATION.  If nothing else, this ensures that MCT_SELF_VALIDATION behaves
// correctly on valid tables.
#undef  MCT_SELF_VALIDATION
#define MCT_SELF_VALIDATION     1

#include "tests/set-common.hpp"
#include "tests/set-parameters.hpp"


BOOST_AUTO_TEST_SUITE (set_main)


// Note that the tests do not look like checking almost anything.  That's because
// 'test_set' validates and compares itself with a known good implementation after each
// modification.  So, most actual checks just happen elsewhere.

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_set_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;

  BOOST_CHECK       (set.empty ());
  BOOST_CHECK_EQUAL (set.size (), 0u);
  BOOST_CHECK       (!set.uses_dynamic_memory ());
  BOOST_CHECK_EQUAL (set.used_memory (), sizeof (set_base_type));
  BOOST_CHECK_EQUAL (set.begin  (), set.end  ());
  BOOST_CHECK_EQUAL (set.cbegin (), set.cend ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_set_explicit_size_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (1000);

  BOOST_CHECK (!set.uses_dynamic_memory ());

  set.insert (data::generate (0));

  BOOST_CHECK    (set.uses_dynamic_memory ());
  BOOST_CHECK_GE (set.bucket_count (), 1000);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_set_copy_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;
  set_type  set2 (set);

  BOOST_CHECK       (set2.empty ());
  BOOST_CHECK_EQUAL (set2.size (), 0u);
  BOOST_CHECK       (!set2.uses_dynamic_memory ());
  BOOST_CHECK_EQUAL (set2.begin  (), set2.end  ());
  BOOST_CHECK_EQUAL (set2.cbegin (), set2.cend ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_set_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;
  set_type  set2;

  set = set2;

  BOOST_CHECK       (set.empty ());
  BOOST_CHECK_EQUAL (set.size (), 0u);
  BOOST_CHECK       (!set.uses_dynamic_memory ());
  BOOST_CHECK_EQUAL (set.begin  (), set.end  ());
  BOOST_CHECK_EQUAL (set.cbegin (), set.cend ());

  set.insert (RANGE (data::values1 ()));
  set = set2;

  // Check the same invariants if assigning over a non-empty set.
  BOOST_CHECK       (set.empty ());
  BOOST_CHECK_EQUAL (set.size (), 0u);
  BOOST_CHECK       (!set.uses_dynamic_memory ());
  BOOST_CHECK_EQUAL (set.begin  (), set.end  ());
  BOOST_CHECK_EQUAL (set.cbegin (), set.cend ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_set_erase_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  set_type  set;

  for (typename vector <type>::const_iterator scan = data::values1 ().begin ();
       scan != data::values1 ().end (); ++scan)
    BOOST_CHECK_EQUAL (set.erase (*scan), 0u);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_set_rehash_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;

  BOOST_CHECK (!set.uses_dynamic_memory ());

  set.rehash (1000);

  BOOST_CHECK    (!set.uses_dynamic_memory ());
  BOOST_CHECK_GE (set.bucket_count (), 1000);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_set_rehash_cleared_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (RANGE (data::values1 ()));
  set.clear ();

  // Hash tables mustn't just drop memory once they become empty, as that would lead to
  // terrible performance in cases where becoming empty and then being filled again is the
  // norm.
  BOOST_CHECK (set.uses_dynamic_memory ());

  set.rehash (2 * set.bucket_count ());

  // However, when asked to resize internal table anyway, there is no point in allocating
  // new array right away if the table is empty.
  BOOST_CHECK (!set.uses_dynamic_memory ());
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_range_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (RANGE (data::values1 ()));
  if (linked)
    assert_identical_order (set, data::values1 ());
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_range_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;
  set_type  set (data::values1 ().begin (), data::values1 ().begin ());
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_initializer_constructor_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int>::set_type  set_type;

  set_type  set1 { 1, 2, 3, 4 };
  set_type  set2 { 1, 3, 1, 3, 2, 2, 2 };

  if (implementation::linked)
    {
      assert_identical_order (set1, vector <int> { 1, 2, 3, 4 });
      assert_identical_order (set2, vector <int> { 1, 3, 1, 3, 2, 2, 2 });
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_empty_initializer_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;
  set_type  set ({ });
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_find_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (RANGE (data::values1 ()));

  for (typename vector <type>::const_iterator scan = data::values1 ().begin ();
       scan != data::values1 ().end (); ++scan)
    BOOST_CHECK (contains (set, *scan));

  for (typename vector <type>::const_iterator scan = data::values2 ().begin ();
       scan != data::values2 ().end (); ++scan)
    {
      BOOST_CHECK_EQUAL (contains (set, *scan),
                         find (RANGE (data::values1 ()), *scan) != range_end (data::values1 ()));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_count_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (RANGE (data::values1 ()));

  for (typename vector <type>::const_iterator scan = data::values1 ().begin ();
       scan != data::values1 ().end (); ++scan)
    BOOST_CHECK_EQUAL (set.count (*scan), 1u);

  for (typename vector <type>::const_iterator scan = data::values2 ().begin ();
       scan != data::values2 ().end (); ++scan)
    {
      BOOST_CHECK_EQUAL (set.count (*scan),
                         (find (RANGE (data::values1 ()), *scan) != range_end (data::values1 ())
                          ? 1u : 0u));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_equal_range_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (RANGE (data::values1 ()));

  for (typename vector <type>::const_iterator scan = data::values1 ().begin ();
       scan != data::values1 ().end (); ++scan)
    {
      pair <const_iterator, const_iterator>  range (set.equal_range (*scan));
      BOOST_CHECK       (set.key_eq () (*range.first, *scan));
      BOOST_CHECK_EQUAL (++range.first, range.second);
    }

  for (typename vector <type>::const_iterator scan = data::values2 ().begin ();
       scan != data::values2 ().end (); ++scan)
    {
      pair <const_iterator, const_iterator>  range (set.equal_range (*scan));

      if (find (RANGE (data::values1 ()), *scan) == range_end (data::values1 ()))
        BOOST_CHECK_EQUAL (range.first, range.second);
      else
        {
          BOOST_CHECK       (set.key_eq () (*range.first, *scan));
          BOOST_CHECK_EQUAL (++range.first, range.second);
        }
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;

  for (typename vector <type>::const_iterator scan = data::values1 ().begin ();
       scan != data::values1 ().end (); ++scan)
    set.insert (*scan);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_range_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;
  set.insert (RANGE (data::values1 ()));
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_moving_1,
                               implementation, test_implementations)
{
  {
    typedef  typename implementation::template resolve <must_not_be_copied>::set_type  set_type;

    // Make it large, so it never rehashes and thus copies elements.
    set_type  set (1000);

    for (int k = 0; k < 100; ++k)
      {
        set.insert (must_not_be_copied (k));
        set.validate_integrity ();
      }
  }

  {
    typedef  typename implementation::template resolve <must_not_be_moved>::set_type  set_type;

    set_type  set (1000);

    for (int k = 0; k < 100; ++k)
      {
        must_not_be_moved  value (k);

        set.insert (value);
        set.validate_integrity ();
      }
  }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_initializers_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int>::set_type     set_type;
  typedef  typename implementation::template resolve <int>::robust_type  robust_type;

  set_type  set { 0, 1, 2 };
  set.insert ({ 3, 4, 5 });

  vector <int>  expected_vector { 0, 1, 2, 3, 4, 5 };
  robust_type   expected (expected_vector.begin (), expected_vector.end ());

  set.validate_integrity ();
  BOOST_CHECK_EQUAL (set, expected);

  if (implementation::linked)
    assert_identical_order (set, expected);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_emplace_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int_wrapper>::set_type  set_type;

  set_type  set;
  for (int k = 0; k < 50; ++k)
    {
      set.emplace (k);
      set.validate_integrity ();
    }

  BOOST_CHECK_EQUAL (set.size (), 50u);
  if (implementation::linked)
    assert_identical_order (set, create_vector (int_wrapper::wrap, set.size ()));
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_insert_erase_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  set_type  set;

  for (typename vector <type>::const_iterator scan = data::values1 ().begin ();
       scan != data::values1 ().end (); ++scan)
    {
      // Simply insert and immediately erase one element.
      set.insert (*scan);
      BOOST_CHECK (set.erase (*scan));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_comparison_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (RANGE (data::values1 ()));
  set_type  set2 (RANGE (data::values1 ()));
  set_type  set3 (RANGE (data::values2 ()));
  set_type  set4 (RANGE (data::values2 ()));

  BOOST_CHECK_EQUAL (set1, set2);
  BOOST_CHECK_EQUAL (set3, set4);
  BOOST_CHECK_NE    (set1, set3);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_copy_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (RANGE (data::values1 ()));
  set_type  set2 (RANGE (data::values2 ()));
  set_type  set3 (set1);
  set_type  set4 (set2);

  BOOST_CHECK_EQUAL (set3, set1);
  BOOST_CHECK_EQUAL (set4, set2);

  if (linked)
    {
      assert_identical_order (set3, set1);
      assert_identical_order (set4, set2);
    }
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_move_constructor_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (move (set_type (RANGE (data::values1 ()))));
  BOOST_CHECK_EQUAL (set, set_type (RANGE (data::values1 ())));
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (RANGE (data::values1 ()));
  set_type  set2;

  BOOST_CHECK_EQUAL (set2.size (), 0u);

  set2 = set1;
  BOOST_CHECK_EQUAL (set2, set1);

  if (linked)
    assert_identical_order (set1, set2);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_self_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (RANGE (data::values1 ()));
  set_type  set2 (RANGE (data::values1 ()));

  BOOST_CHECK_EQUAL (set1, set2);
  if (linked)
    assert_identical_order (set1, set2);

  set1 = set1;

  BOOST_CHECK_EQUAL (set1, set2);
  if (linked)
    assert_identical_order (set1, set2);
}


#if MCT_CXX0X_SUPPORTED

BOOST_AUTO_TEST_CASE_TEMPLATE (test_move_assignment_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set;

  set = move (set_type (RANGE (data::values1 ())));
  BOOST_CHECK_EQUAL (set, set_type (RANGE (data::values1 ())));
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_initializer_assignment_1,
                               implementation, test_implementations)
{
  typedef  typename implementation::template resolve <int>::set_type  set_type;

  set_type  set { 1, 2, 3, 4 };
  set = { 4, 5, 6 };

  if (implementation::linked)
    assert_identical_order (set, vector <int> { 4, 5, 6 });
}

#endif  // MCT_CXX0X_SUPPORTED


BOOST_AUTO_TEST_CASE_TEMPLATE (test_swap_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set1 (RANGE (data::values1 ()));
  set_type  set2 (RANGE (data::values2 ()));
  set_type  set3 (RANGE (data::values1 ()));
  set_type  set4 (RANGE (data::values2 ()));

  BOOST_CHECK_EQUAL (set1, set3);
  BOOST_CHECK_EQUAL (set2, set4);

  if (linked)
    {
      assert_identical_order (set1, set3);
      assert_identical_order (set2, set4);
    }

  swap (set3, set4);

  BOOST_CHECK_EQUAL (set1, set4);
  BOOST_CHECK_EQUAL (set2, set3);

  if (linked)
    {
      assert_identical_order (set1, set4);
      assert_identical_order (set2, set3);
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_clear_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  set_type  set (RANGE (data::values1 ()));
  set.clear ();
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_value_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  set_type  set1 (RANGE (data::values1 ()));
  set_type  set2 (range_begin (data::values1 ()) + 1, range_end (data::values1 ()));

  // Make sure funny comparison functions don't ruin this test.
  if (set1.size () != set2.size ())
    {
      BOOST_CHECK       (set1.erase (data::values1 () [0]));
      BOOST_CHECK_EQUAL (set1, set2);

      if (linked)
        assert_identical_order (set1, set2);
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_position_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  set_type  set (RANGE (data::values1 ()));

  const size_t  num_passes_expected = set.size ();
  size_t        num_passes          = 0;

  for (iterator scan = set.begin (), end = set.end (); scan != end; ++num_passes)
    scan = set.erase (scan);

  BOOST_CHECK       (set.empty ());
  BOOST_CHECK_EQUAL (num_passes, num_passes_expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_erase_range_1,
                               parameters, test_parameters_normal)
{
  COMMON_TEST_SETUP;

  set_type  set (RANGE (data::values1 ()));

  set.erase (set.begin (), set.end ());
  BOOST_CHECK (set.empty ());
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_growth_1,
                               parameters, test_parameters_std_equal)
{
  COMMON_TEST_SETUP;

  // Note: fairly small, because testing after each step is slow.
  set_type  set;
  for (int k = 0; k < 200; ++k)
    set.insert (data::generate (k));

  BOOST_CHECK_EQUAL (set.size (), 200u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, set.size ()));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_shrink_1,
                               parameters, test_parameters_std_equal_normal)
{
  COMMON_TEST_SETUP;

  const size_t   num_items = 200;
  vector <type>  items (create_vector (data::generate, num_items));
  set_type       set   (items.begin (), items.end ());

  BOOST_CHECK_EQUAL (set.size (), num_items);

  const_iterator  almost_end = set.end ();
  for (int k = 0; k < 10; ++k)
    --almost_end;

  set.erase (set.begin (), almost_end);

  BOOST_CHECK_EQUAL (set.size (), 10u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, num_items, num_items - 10));

  set.rehash (32);

  BOOST_CHECK_EQUAL (set.bucket_count (), 32u);
  if (linked)
    assert_identical_order (set, create_vector (data::generate, num_items, num_items - 10));
}

BOOST_AUTO_TEST_CASE_TEMPLATE (test_tiny_set_1,
                               parameters, test_parameters)
{
  COMMON_TEST_SETUP;

  static  const float  load_factors[] = { 0.01, 0.1, 0.25, 0.5, 0.75, 0.9, 0.99 };

  for (const float* scan = range_begin (load_factors); scan != range_end (load_factors); ++scan)
    {
      BOOST_TEST_CHECKPOINT ("testing with maximum load factor " << *scan);

      set_type  set (1);
      set.max_load_factor (*scan);

      for (int k = 0; k < 50; ++k)
        set.insert (data::generate (k));

      if (linked)
        assert_identical_order (set, create_vector (data::generate, 50));
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_max_load_factor_decrease_1,
                               implementation, test_implementations)
{
  typedef  test_set <typename implementation::template resolve <int>::set_type>  set_type;

  set_type  set;

  // Make sure 'set' is full for the default maximum load factor.
  for (int k = 0; ; k++)
    {
      set_type  copy (set);
      copy.insert (k);

      if (copy.bucket_count () == set.bucket_count ())
        set.swap (copy);
      else
        break;
    }

  // Just something very small; implementation will clamp to an acceptable value.
  set.max_load_factor (0.0f);
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_intrusive_bucket_size_1,
                               implementation, test_implementations)
{
  typedef  test_set <typename implementation::template resolve <ham>::set_type>  set_type;

  set_type  set;

  BOOST_CHECK       (!set_type::bucket_type::KEEPS_HASHES);
  BOOST_CHECK_EQUAL (sizeof (typename set_type::iterator), sizeof (ham*));
}


BOOST_AUTO_TEST_CASE_TEMPLATE (test_cross_class_comparison_1,
                               implementation, test_implementations)
{
  typedef  test_set <typename implementation::template resolve <int>::set_type>  set_type;

  vector <int>  values1 (create_vector (Data <int>::generate, 10));
  vector <int>  values2 (values1);

  values2[3] = 100;

  set_type  set1 (RANGE (values1));
  set_type  set2 (RANGE (values2));

  BOOST_CHECK_EQUAL (set1, closed_hash_set <int> (RANGE (values1)));
  BOOST_CHECK_NE    (set1, closed_hash_set <int> (RANGE (values2)));

  BOOST_CHECK_EQUAL (set1, linked_hash_set <int> (RANGE (values1)));
  BOOST_CHECK_NE    (set1, linked_hash_set <int> (RANGE (values2)));

  BOOST_CHECK_EQUAL (set1, huge_linked_hash_set <int> (RANGE (values1)));
  BOOST_CHECK_NE    (set1, huge_linked_hash_set <int> (RANGE (values2)));

  BOOST_CHECK_EQUAL (set1, forward_hash_set <int> (RANGE (values1)));
  BOOST_CHECK_NE    (set1, forward_hash_set <int> (RANGE (values2)));

  BOOST_CHECK_EQUAL (set1, huge_forward_hash_set <int> (RANGE (values1)));
  BOOST_CHECK_NE    (set1, huge_forward_hash_set <int> (RANGE (values2)));
}


BOOST_AUTO_TEST_SUITE_END ()


// Local variables:
// mode: c++
// c-basic-offset: 2
// indent-tabs-mode: nil
// fill-column: 90
// End:
